/* LoRa Sender Node
 *
 * Function:
 * Assemble packets using the PacketHandler and forward data to the nominated Gateway Node
 * 
 * Description:
 * 1. Uses packet structures defined in the PacketHandler library
 * 2. Output generated for Serial Monitor and/or OLED diaplay
 *
 * 10 Dec 2022 v1.0 Code sample for Heltec LoRa WiFi 32 V3
 * 
 * 10 Dec 2022
 * Digital Concepts
 * www.digitalconcepts.net.au
 */

#include "Arduino.h"
#include "LoRaWan_APP.h"
#include <SSD1306.h>            // OLED display
#include <PacketHandler.h>      // [Langlo] LoRa Packet management
#include <OneWire.h>            // Required for Dallas Temperature
#include <DallasTemperature.h>  // DS18B20 Library
#include "Adafruit_BME280.h"    // BME280 class & methods
#include "Adafruit_Sensor.h"    // BME280 sensor support methods
#include "Wire.h"               // Need this to define I2C bus

#include <Ticker.h>             // Asynchronous task management

#include "LangloLoRa.h"       // LoRa configuration parameters
#include <WiFiLoRa32V3Pins.h>   // Heltec WiFi LoRa 32 V3 pin definitions
#include <nodeV3X1Config.h>     // Specific local Node configuration parameters

#define batteryMonitor    A4a_A0
#define wakePin           A4a_A1

Ticker oledTicker;

const int reportingPeriod = 60000;
const int oledActivationInterval = 30; // 300 seconds (5 min) OLED Auto-off (Set to 0 for always on)
const int oledControlPin =  GPIO0;     // Hardware interrupt
bool oledActive = true;
bool oledOff = true;

SSD1306 myDisplay(0x3c, SDA_OLED, SCL_OLED), *displayPtr;

// DS18B20
const int oneWireBus = A4a_A2;
const int temperaturePrecision = 9;
int nodeTemperature = 0;

// Setup a oneWire instance to communicate with any OneWire devices (not just Maxim/Dallas temperature ICs)
OneWire oneWire(oneWireBus);

// Pass our oneWire reference to Dallas Temperature.
DallasTemperature ds18b20Sensors(&oneWire);

// BME280 Settings
const int I2C_BME_Address = 0x76;

// Define BME280 sensor
Adafruit_BME280 bme280;             // Defaults (I2C_BME_Address = 0x77) set within Adafruit libraries

// Define the second I2C bus for the BME280
TwoWire sensorI2C = TwoWire(1);

static RadioEvents_t RadioEvents;
void OnTxDone( void );
void OnTxTimeout( void );

// LowPower stuff
// const int reportInterval = 60000;   // Report values every 60000 milliseconds (60 seconds)

#define wakeTime 1000
#define sleepTime 60000
//static TimerEvent_t sleep;
//static TimerEvent_t wakeUp;

typedef enum
{
    LOWPOWER,
    ReadVoltage,
    TX_State
} States_t;

States_t state;
bool sleepMode = false;
int16_t rssi,rxSize;
uint16_t batteryVoltage;

// Instantiate the Packet Handler
PacketHandler packet;

// Sensor value variables
uint16_t humidity = 0;
uint16_t pressure = 0;
int temperature = 0;

const float LSB = 0.02761647; // per manual calibration (when using int in 1/10mm)
uint8_t inputRainfall = 0;
int adc0 = 0;

uint16_t messageCounter = 0;

void onSleep()
{
  Serial.printf("[onSleep] Go into LowPower mode for %d ms\r\n",sleepTime);
  state = LOWPOWER;
//  TimerSetValue( &wakeUp, sleepTime );
//  TimerStart( &wakeUp );
}

void onWake()
{
  Serial.println(""); 
  Serial.printf("[onWake] Wake for %d ms, then go back into LowPower mode.\r\n",wakeTime);
  state = ReadVoltage;
//  TimerSetValue( &sleep, wakeTime );
//  TimerStart( &sleep );
}

void setup() {
  Serial.begin(115200);
  Mcu.begin();
  delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" LoRa/Display Test");
  Serial.println();
  Serial.println("[setup] Commence Set-up...");
  
  pinMode(Builtin_LED,OUTPUT);
  pinMode(Vext_Ctrl,OUTPUT);
  pinMode(oledControlPin,INPUT_PULLUP);     // Activate OLED display
  pinMode(wakePin,OUTPUT);                  // Activate battery monitor

// Initialise the OLED display
  
  Serial.println("[setup] Initialise display...");
  pinMode(RST_OLED,OUTPUT);     // GPIO16
  digitalWrite(RST_OLED,LOW);   // set GPIO16 low to reset OLED
  delay(50);
  digitalWrite(RST_OLED,HIGH);

  myDisplay.init();
  myDisplay.clear();
//  myDisplay.flipScreenVertically();
  myDisplay.setFont(ArialMT_Plain_10);
  myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
  myDisplay.drawString(5,5,"LoRa Sender");
  myDisplay.display();
  oledOff = false;
  displayPtr = &myDisplay;

  batteryVoltage=0;
  rssi=0;

//  Initialise LoRa

  Serial.println("[setup] Initialise LoRa...");
  Serial.println("[setup] Activate LoRa Receiver");
    
  // LoRa parameters defined in LangloLoRa.h
  
  RadioEvents.TxDone = OnTxDone;
 
  Radio.Init( &RadioEvents );
  Radio.SetChannel( RF_FREQUENCY );
  Radio.SetTxConfig( MODEM_LORA, TX_OUTPUT_POWER, 0, LORA_BANDWIDTH,
                                 LORA_SPREADING_FACTOR, LORA_CODINGRATE,
                                 LORA_PREAMBLE_LENGTH, LORA_FIX_LENGTH_PAYLOAD_ON,
                                 true, 0, 0, LORA_IQ_INVERSION_ON, 3000 );

  Radio.Sleep( );

  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();
  
//  TimerInit( &sleep, onSleep );
//  TimerInit( &wakeUp, onWake );
  onWake();
  
  // initialise the Packet Handler
  
  packet.begin(gatewayMAC,myMAC);

  if ( oledActivationInterval > 0 ) {
    oledTicker.once( oledActivationInterval, deactivateOLED );
    attachInterrupt( oledControlPin, activateOLED, FALLING);
  }

  Serial.println("[setup] Set-up Complete");
  Serial.println(); 
}

void loop()
{
	switch(state)
	{
		case TX_State:
    {
      Serial.println("[loop] Send Messages");

      if ( oledActive ) {
        if ( oledOff ) {
          Serial.println();
          Serial.println("[loop] Turn on OLED display...");
          Serial.println();
          myDisplay.displayOn(); // Switch on display
          delay(10);
          oledOff = false;
          myDisplay.clear();
          myDisplay.setFont(ArialMT_Plain_10);
          myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
          myDisplay.drawString(5,5,"LoRa [Gateway] Receiver");
          myDisplay.display();       
          oledTicker.once( oledActivationInterval, deactivateOLED ); // Start the timer
        }
      }

      // Populate the packet
    
      packet.setSequenceNumber(messageCounter);
      packet.setPacketType(VOLTAGE);

      packet.setVoltage(batteryVoltage);
      packet.hexDump();
      packet.serialOut();
      if ( oledActive ) packet.displayOut(displayPtr);  
      
      digitalWrite(Builtin_LED,HIGH);
      sendMessage();
      digitalWrite(Builtin_LED,LOW);
      
      // Have a snooze while the gateway hadndles that one
      delay(100);  //LowPower time

      packet.setPacketType(TEMPERATURE);
      readDS18B20Sensors();
      packet.setTemperature(nodeTemperature);
      packet.hexDump();
      packet.serialOut();   
      if ( oledActive ) packet.displayOut(displayPtr);  
            
      digitalWrite(Builtin_LED,HIGH);
      sendMessage();
      digitalWrite(Builtin_LED,LOW);
      
      // Have a snooze while the gateway hadndles that one
      delay(100);  //LowPower time

      packet.setPacketType(ATMOSPHERE);
      readBMESensor();
      packet.setTemperature(temperature);
      packet.setPressure(pressure);
      packet.setHumidity(humidity);
      packet.hexDump();
      packet.serialOut();   
      if ( oledActive ) packet.displayOut(displayPtr);  
            
      digitalWrite(Builtin_LED,HIGH);
      sendMessage();
      digitalWrite(Builtin_LED,LOW);
      
      Serial.println("[loop] Increment Message Counter");
      messageCounter++;                 // Increment message count
      Serial.println(""); 
      state=LOWPOWER;
      break;
    }
		case LOWPOWER:
		{
//      lowPowerHandler();
      delay(reportingPeriod);
      onWake();
      break;
		}
    case ReadVoltage:
    {
      batteryVoltage=readBatteryVoltage();
      Serial.print("[readSensor] Battery Voltage: ");
      Serial.print( batteryVoltage );
      Serial.println(" mV");
/*      
      display.clear();
      display.setFont(ArialMT_Plain_10);
      display.drawString(10, 5, "Voltage:");
      display.setFont(ArialMT_Plain_16);
      String voltageString = String(batteryVoltage);
      voltageString.concat(" mV");
      display.drawString(10, 20, voltageString);
      display.display();
*/
      state = TX_State;
      break;
    }
    default:
        break;
	}
//    Radio.IrqProcess( );
}

void OnTxDone( void )
{
  Serial.println("[OnTxDone] TX done!");
  state=LOWPOWER;
}

void OnTxTimeout( void )
{
  Radio.Sleep( );
  Serial.println("[OnTxTimeout] TX Timeout...");
  state=LOWPOWER;
}

uint16_t readBatteryVoltage() {
  static const float voltageFactor = (3.2/4096)*(133.0/100)*(4.14/4.20); // (max. ADC voltage/ADC range) * voltage divider factor * calibration factor
  uint16_t inputValue = 0, peakValue = 0;
  Serial.println("[readBatteryLevel] Read ADC...");
  
  digitalWrite(wakePin,HIGH);
  delay(900);
  for (int i = 0; i < 10; i++) {
    delay(10);
    inputValue = analogRead(batteryMonitor); // and read ADC (Note: 33kΩ/100kΩ voltage divider in monitoring circuit)
    Serial.print(String(inputValue) + " = ");
    Serial.println(String(inputValue*voltageFactor) + "V");
    if (inputValue > peakValue) peakValue = inputValue;
  }
  Serial.print(String(peakValue) + " = ");
  Serial.println(String(peakValue*voltageFactor) + "V");
  Serial.println();
  digitalWrite(wakePin,LOW);
  
  uint16_t voltage = (int)(peakValue * voltageFactor * 1000);
  Serial.print("[readBAtteryVoltage] Returned Value: ");
  Serial.println(voltage);

  return voltage;
}

void readDS18B20Sensors() {
  Serial.println("[readDS18B20Sensors] Read sensors...");
/*
 * If this node is reading atmospheric conditions from a BME sensor
 * All values recorded as integers, temperature multiplied by 10 to keep one decimal place
*/
  digitalWrite(Vext_Ctrl, LOW);    // Make sure the external power supply is on
  delay(10);
  ds18b20Sensors.begin();
  delay(100);                 // Give everything a moment to settle down
  
  ds18b20Sensors.requestTemperatures(); // Send the command to get temperatures
  // After we got the temperatures, we can print them here.
  // We use the function ByIndex, and as an example get the temperature from the first sensor only.
  Serial.println("[readDS18B20Sensors] Device 1 (index 0)");

  float tempTemperature = ds18b20Sensors.getTempCByIndex(0);
  Serial.print("[readDS18B20Sensors] Returned value: ");
  Serial.println(tempTemperature);

  nodeTemperature = (int) (10*tempTemperature);
 
  Serial.print("[readDS18B20Sensors] Temperature: ");
  Serial.println((float) nodeTemperature/10, 1);
}

void readBMESensor() {
  Serial.println("[readBMESensor] Read sensor...");
/*
 * If this node is reading atmospheric conditions from a BME sensor
 * All values recorded as integers, temperature multiplied by 10 to keep one decimal place
*/
  digitalWrite(Vext_Ctrl, LOW);      // Make sure the external power supply is on
  delay(10);                         // Give everything a moment to settle down
  sensorI2C.begin(SDA,SCL,400000);
  if (bme280.begin(I2C_BME_Address, &sensorI2C)) {
    Serial.println( "[readBMESensor] BME sensor initialisation complete" );
  } else {
    Serial.println( "[readBMESensor] BME sensor initialisation error" );
  }

  delay(100); // The BME280 needs a moment to get itself together... (50ms is too little time)
 
  temperature = (int) (10*bme280.readTemperature());
  pressure = (uint16_t) (bme280.readPressure() / 91.79F);
  humidity = (uint16_t) (bme280.readHumidity());

  sensorI2C.end();

  Serial.println("");
  Serial.print("[readBMESensor] Temperature: ");
  Serial.println((float) temperature/10, 1);
  Serial.print("[readBMESensor] Pressure: ");
  Serial.println(pressure);
  Serial.print("[readBMESensor] Humidity: ");
  Serial.println(humidity);
}

void activateOLED() {
 /* 
  Serial.println();
  Serial.println("[activateOLED] Interrupt intercepted... activating OLED");
  Serial.println();
  
  myDisplay.displayOn(); // Switch on display
  
  digitalWrite(RST_OLED,LOW);  // set GPIO16 LOW to reset OLED
  delay(50);
  digitalWrite(RST_OLED,HIGH);

  myDisplay.init();
  myDisplay.clear();
//  myDisplay.flipScreenVertically();
  myDisplay.setFont(ArialMT_Plain_10);

  myDisplay.setTextAlignment(TEXT_ALIGN_LEFT);
  myDisplay.drawString(5,5,"LoRa [Gateway] Receiver");
  myDisplay.display();

  oledTicker.once( oledActivationInterval, deactivateOLED );
*/
  oledActive = true;
}

void deactivateOLED() {
  Serial.println();
  Serial.println("[deactivateOLED] Timer expired... deactivating OLED");
  Serial.println();

  myDisplay.displayOff(); // Switch off display
  oledActive = false;
  oledOff = true;
}

void sendMessage() {
  Serial.println("[sendMessage] Finished Packet");
  Radio.Send(packet.byteStream(), packet.packetByteCount());
  Serial.println("[sendMessage] Packet Sent");
  Serial.println();
}